/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.example.server;

import ca.uhn.fhir.rest.server.exceptions.InvalidRequestException;
import ca.uhn.hapi.fhir.cdshooks.api.CdsService;
import ca.uhn.hapi.fhir.cdshooks.api.CdsServiceFeedback;
import ca.uhn.hapi.fhir.cdshooks.api.CdsServicePrefetch;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceAcceptedSuggestionJson;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceFeedbackJson;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceIndicatorEnum;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceRequestJson;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceResponseCardJson;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceResponseCardSourceJson;
import ca.uhn.hapi.fhir.cdshooks.api.json.CdsServiceResponseJson;
import com.example.server.custom.extension.model.MeasurementUnitExtension;
import org.apache.commons.lang3.StringUtils;
import org.hl7.fhir.r4.model.HumanName;
import org.hl7.fhir.r4.model.Patient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Instant;
import java.time.LocalDate;
import java.time.Period;
import java.time.ZoneId;
import java.util.Date;
import java.util.List;
import java.util.UUID;

public class AgeCalculatorCdsService {

	public static final String SERVICE_ID = "calculate-age";
	public static final String PATIENT_KEY = "patient";
	private static final String MONTHS = "months";
	private static final String YEARS = "years";
	private static final Logger ourLog = LoggerFactory.getLogger(AgeCalculatorCdsService.class);

	// START SNIPPET: ageCalculatorCdsService
	@CdsService(
			value = SERVICE_ID,
			hook = "patient-view",
			title = "Age",
			description = "Age calculated from patient birthday",
			prefetch = {@CdsServicePrefetch(value = "patient", query = "Patient/{{context.patientId}}")},
			allowAutoFhirClientPrefetch = true,
			extension = """
				{
					"age-measurement-unit": "years or months"
				}
				""",
			extensionClass = MeasurementUnitExtension.class)
	public CdsServiceResponseJson calculateAge(CdsServiceRequestJson theCdsServiceRequestJson) {
		Patient patient = (Patient) theCdsServiceRequestJson.getPrefetch(PATIENT_KEY);

		if (patient == null) {
			throw new InvalidRequestException("'patient' prefetch resource missing from request.");
		}

		String ageMeasurementUnit = YEARS;
		// Get the extension passed in the request
		MeasurementUnitExtension extension = (MeasurementUnitExtension) theCdsServiceRequestJson.getExtension();
		if (!StringUtils.isEmpty(extension.getAgeMeasurementUnit())) {
			ageMeasurementUnit = (MONTHS).equals(extension.getAgeMeasurementUnit()) ? MONTHS : YEARS;
		}

		CdsServiceResponseJson response = new CdsServiceResponseJson();
		CdsServiceResponseCardJson card = new CdsServiceResponseCardJson();
		Integer ageInRequestedMeasurement = calculateAge(patient, ageMeasurementUnit);
		String name = getName(patient);

		if (ageInRequestedMeasurement == null) {
			card.setSummary(name + "'s age is unknown");
		} else {
			card.setSummary(name + " is " + ageInRequestedMeasurement + " " + ageMeasurementUnit + " old.");
		}
		card.setIndicator(CdsServiceIndicatorEnum.INFO);
		CdsServiceResponseCardSourceJson source = new CdsServiceResponseCardSourceJson();
		source.setLabel("ACME Age Calculator");
		card.setSource(source);
		response.addCard(card);
		return response;
	}

	private String getName(Patient thePatient) {
		if (thePatient == null) {
			return null;
		}
		List<HumanName> names = thePatient.getName();
		if (names == null || names.size() == 0 || names.get(0).isEmpty()) {
			return "This patient";
		}
		HumanName nameItem = names.get(0);
		return nameItem.getNameAsSingleString();
	}

	private Integer calculateAge(Patient thePatient, String theAgeMeasurementUnit) {
		if (thePatient == null) {
			return null;
		}
		Date birthDate = thePatient.getBirthDate();
		if (birthDate == null) {
			return null;
		}
		LocalDate localBirthDate =
				birthDate.toInstant().atZone(ZoneId.systemDefault()).toLocalDate();
		int ageDifferenceInYears = Period.between(
						localBirthDate,
						Instant.now().atZone(ZoneId.systemDefault()).toLocalDate())
				.getYears();
		return theAgeMeasurementUnit.equals(MONTHS) ? ageDifferenceInYears * 12 : ageDifferenceInYears;
	}

	// END SNIPPET: ageCalculatorCdsService

	@CdsServiceFeedback("calculate-age")
	public CdsServiceFeedbackJson feedback(CdsServiceFeedbackJson theFeedback) {
		ourLog.info("Feedback for {} recieved at {}", theFeedback.getCard(), theFeedback.getOutcomeTimestamp());
		theFeedback.setAcceptedSuggestions(List.of(
				new CdsServiceAcceptedSuggestionJson().setId(UUID.randomUUID().toString())));
		return theFeedback;
	}
}
